# Control Page

{{ control_writer.write_control_with_sections(control, ['statement', 'objective', 'expected_evidence', 'implementation_guidance', 'table_of_parameters'], {'statement':'Control Statement Header', 'objective':'Control Objective Header', 'expected_evidence':'Expected Evidence Header', 'implementation_guidance':'Implementation Guidance', 'table_of_parameters':'Table of Control Parameters'}) }}
